@@ -20,6 +20,7 @@
  ***************************************************************************/
 
 #include "PackageKitBackend.h"
+#include "PackageKitSourcesBackend.h"
 #include "PackageKitResource.h"
 #include "PackageKitUpdater.h"
 #include "AppPackageKitResource.h"
@@ -27,20 +28,34 @@
 #include "AppstreamReviews.h"
 #include <resources/AbstractResource.h>
 #include <resources/StandardBackendUpdater.h>
+#include <resources/SourcesModel.h>
 #include <Transaction/TransactionModel.h>
+
+#include <QProcess>
 #include <QStringList>
 #include <QDebug>
 #include <QTimer>
 #include <QTimerEvent>
+#include <QStandardPaths>
+
 #include <PackageKit/Transaction>
 #include <PackageKit/Daemon>
 #include <PackageKit/Details>
 
+#include <KDesktopFile>
 #include <KLocalizedString>
 #include <QAction>
 
+#include "utils.h"
+#include "config-paths.h"
+
 MUON_BACKEND_PLUGIN(PackageKitBackend)
 
+static QString locateService(const QString &filename)
+{
+    return QStandardPaths::locate(QStandardPaths::GenericDataLocation, QStringLiteral("applications/")+filename);
+}
+
 PackageKitBackend::PackageKitBackend(QObject* parent)
     : AbstractResourcesBackend(parent)
     , m_updater(new PackageKitUpdater(this))
@@ -48,10 +63,19 @@ PackageKitBackend::PackageKitBackend(QObject* parent)
     , m_isFetching(0)
     , m_reviews(new AppstreamReviews(this))
 {
+#ifdef NEWAPPSTREAM
+    bool b = m_appdata.load();
+#else
     bool b = m_appdata.open();
-    Q_ASSERT(b && "must be able to open the appstream database");
+#endif
     if (!b) {
-        qWarning() << "Couldn't open the AppStream database";
+        qWarning() << "Could not open the AppStream metadata pool";
+
+        auto msg = new QAction(i18n("Got it"), this);
+        msg->setWhatsThis(i18n("Please make sure that Appstream is properly set up on your system"));
+        msg->setPriority(QAction::HighPriority);
+        connect(msg, &QAction::triggered, msg, [msg](){ msg->setVisible(false); });
+        m_messageActions << msg;
     }
     reloadPackageList();
 
@@ -61,21 +85,49 @@ PackageKitBackend::PackageKitBackend(QObject* parent)
     t->setSingleShot(false);
     t->start();
 
+    m_delayedDetailsFetch.setSingleShot(true);
+    m_delayedDetailsFetch.setInterval(0);
+    connect(&m_delayedDetailsFetch, &QTimer::timeout, this, &PackageKitBackend::performDetailsFetch);
+
     QAction* updateAction = new QAction(this);
     updateAction->setIcon(QIcon::fromTheme(QStringLiteral("system-software-update")));
     updateAction->setText(i18nc("@action Checks the Internet for updates", "Check for Updates"));
     updateAction->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_R));
+    connect(this, &PackageKitBackend::fetchingChanged, updateAction, [updateAction, this](){
+        updateAction->setEnabled(!isFetching());
+    });
     connect(updateAction, &QAction::triggered, this, &PackageKitBackend::refreshDatabase);
     m_messageActions += updateAction;
 
+    const auto service = locateService(QStringLiteral("software-properties-kde.desktop"));
+    if (!service.isEmpty())
+        m_messageActions += createActionForService(service);
+
     connect(PackageKit::Daemon::global(), &PackageKit::Daemon::updatesChanged, this, &PackageKitBackend::fetchUpdates);
     connect(PackageKit::Daemon::global(), &PackageKit::Daemon::isRunningChanged, this, &PackageKitBackend::checkDaemonRunning);
+    connect(m_reviews, &AppstreamReviews::ratingsReady, this, &AbstractResourcesBackend::emitRatingsReady);
+
+    SourcesModel::global()->addSourcesBackend(new PackageKitSourcesBackend(this));
 }
 
 PackageKitBackend::~PackageKitBackend()
 {
 }
 
+QAction* PackageKitBackend::createActionForService(const QString &servicePath)
+{
+    QAction* action = new QAction(this);
+    KDesktopFile parser(servicePath);
+    action->setIcon(QIcon::fromTheme(parser.readIcon()));
+    action->setText(parser.readName());
+    connect(action, &QAction::triggered, action, [servicePath, this](){
+        bool b = QProcess::startDetached(QStringLiteral(CMAKE_INSTALL_FULL_LIBEXECDIR_KF5 "/discover/runservice"), {servicePath});
+        if (!b)
+            qWarning() << "Could not start" << servicePath;
+    });
+    return action;
+}
+
 bool PackageKitBackend::isFetching() const
 {
     return m_isFetching;
@@ -96,76 +148,160 @@ void PackageKitBackend::acquireFetching(bool f)
 
 void PackageKitBackend::reloadPackageList()
 {
-    m_updatingPackages = m_packages;
-    
+    acquireFetching(true);
     if (m_refresher) {
         disconnect(m_refresher.data(), &PackageKit::Transaction::finished, this, &PackageKitBackend::reloadPackageList);
     }
 
-    foreach(const Appstream::Component& component, m_appdata.allComponents()) {
-        const auto res = new AppPackageKitResource(component, this);;
-        m_updatingPackages.packages[component.id()] = res;
+#ifdef NEWAPPSTREAM
+    const auto components = m_appdata.components();
+#else
+    const auto components = m_appdata.allComponents();
+#endif
+    QStringList neededPackages;
+    neededPackages.reserve(components.size());
+    foreach(const Appstream::Component& component, components) {
+        if (component.packageNames().isEmpty()) {
+            qDebug() << "no packages for" << component.name();
+            continue;
+        }
+        neededPackages += component.packageNames();
+
+        const auto res = new AppPackageKitResource(component, this);
+        m_packages.packages[component.id()] = res;
         foreach (const QString& pkg, component.packageNames()) {
-            m_updatingPackages.packageToApp[pkg] += component.id();
+            m_packages.packageToApp[pkg] += component.id();
         }
 
         foreach (const QString& pkg, component.extends()) {
-            m_updatingPackages.extendedBy[pkg] += res;
+            m_packages.extendedBy[pkg] += res;
         }
     }
+    acquireFetching(false);
+    neededPackages.removeDuplicates();
 
-    PackageKit::Transaction * t = PackageKit::Daemon::getPackages();
-    connect(t, &PackageKit::Transaction::finished, this, &PackageKitBackend::getPackagesFinished);
-    connect(t, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackage);
-    connect(t, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
-    acquireFetching(true);
+    qDebug() << "needed..." << neededPackages.count();
+    resolvePackages(neededPackages);
+}
 
+class TransactionSet : public QObject
+{
+    Q_OBJECT
+    public:
+        TransactionSet(const QVector<PackageKit::Transaction*> &transactions)
+            : m_transactions(transactions)
+        {
+            foreach(PackageKit::Transaction* t, transactions) {
+                connect(t, &PackageKit::Transaction::finished, this, &TransactionSet::transactionFinished);
+            }
+        }
+
+        void transactionFinished(PackageKit::Transaction::Exit exit) {
+            PackageKit::Transaction* t = qobject_cast<PackageKit::Transaction*>(sender());
+            if (exit != PackageKit::Transaction::ExitSuccess) {
+                qWarning() << "failed" << exit << t;
+            }
+
+            m_transactions.removeAll(t);
+            if (m_transactions.isEmpty()) {
+                Q_EMIT allFinished();
+            }
+        }
+
+    Q_SIGNALS:
+        void allFinished();
+
+    private:
+        QVector<PackageKit::Transaction*> m_transactions;
+
+};
+
+void PackageKitBackend::clearPackages(const QStringList& packageNames)
+{
+    const auto resources = resourcesByPackageNames<QVector<AbstractResource*>>(packageNames);
+    for(auto res: resources) {
+        qobject_cast<PackageKitResource*>(res)->clearPackageIds();
+    }
+}
+
+void PackageKitBackend::resolvePackages(const QStringList &packageNames)
+{
+    PackageKit::Transaction * tArch = PackageKit::Daemon::resolve(packageNames, PackageKit::Transaction::FilterArch);
+    connect(tArch, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackageArch);
+    connect(tArch, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
+
+    PackageKit::Transaction * tNotArch = PackageKit::Daemon::resolve(packageNames, PackageKit::Transaction::FilterNotArch);
+    connect(tNotArch, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackageNotArch);
+    connect(tNotArch, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
+
+    TransactionSet* merge = new TransactionSet({tArch, tNotArch});
+    connect(merge, &TransactionSet::allFinished, this, &PackageKitBackend::getPackagesFinished);
     fetchUpdates();
 }
 
 void PackageKitBackend::fetchUpdates()
 {
-    m_updatesPackageId.clear();
-
     PackageKit::Transaction * tUpdates = PackageKit::Daemon::getUpdates();
     connect(tUpdates, &PackageKit::Transaction::finished, this, &PackageKitBackend::getUpdatesFinished);
     connect(tUpdates, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackageToUpdate);
     connect(tUpdates, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
-    acquireFetching(true);
+    m_updatesPackageId.clear();
+
+    m_updater->setProgressing(true);
 }
 
+void PackageKitBackend::addPackageArch(PackageKit::Transaction::Info info, const QString& packageId, const QString& summary)
+{
+    addPackage(info, packageId, summary, true);
+}
 
-void PackageKitBackend::addPackage(PackageKit::Transaction::Info info, const QString &packageId, const QString &summary)
+void PackageKitBackend::addPackageNotArch(PackageKit::Transaction::Info info, const QString& packageId, const QString& summary)
+{
+    addPackage(info, packageId, summary, false);
+}
+
+void PackageKitBackend::addPackage(PackageKit::Transaction::Info info, const QString &packageId, const QString &summary, bool arch)
 {
     const QString packageName = PackageKit::Daemon::packageName(packageId);
-    QVector<AbstractResource*> r = resourcesByPackageName(packageName, true);
+    QSet<AbstractResource*> r = resourcesByPackageName(packageName);
     if (r.isEmpty()) {
-        r += new PackageKitResource(packageName, summary, this);
-        m_updatingPackages.packages[packageName] = r.last();
+        auto pk = new PackageKitResource(packageName, summary, this);
+        r = { pk };
+        m_packagesToAdd.insert(pk);
     }
     foreach(auto res, r)
-        static_cast<PackageKitResource*>(res)->addPackageId(info, packageId, summary);
+        static_cast<PackageKitResource*>(res)->addPackageId(info, packageId, arch);
 }
 
-void PackageKitBackend::getPackagesFinished(PackageKit::Transaction::Exit exit)
+void PackageKitBackend::getPackagesFinished()
 {
-    Q_ASSERT(m_isFetching);
-
-    if (exit != PackageKit::Transaction::ExitSuccess) {
-        qWarning() << "error while fetching details" << exit;
-    }
-
-    for(auto it = m_updatingPackages.packages.begin(); it != m_updatingPackages.packages.end(); ) {
+    for(auto it = m_packages.packages.cbegin(); it != m_packages.packages.cend(); ++it) {
         auto pkr = qobject_cast<PackageKitResource*>(it.value());
         if (pkr->packages().isEmpty()) {
             qWarning() << "Failed to find package for" << it.key();
-            it.value()->deleteLater();
-            it = m_updatingPackages.packages.erase(it);
-        } else
-            ++it;
+            m_packagesToDelete += pkr;
+        }
     }
+    includePackagesToAdd();
+}
 
-    m_packages = m_updatingPackages;
+void PackageKitBackend::includePackagesToAdd()
+{
+    if (m_packagesToAdd.isEmpty() && m_packagesToDelete.isEmpty())
+        return;
+
+    acquireFetching(true);
+    foreach(PackageKitResource* res, m_packagesToAdd) {
+        m_packages.packages[res->packageName()] = res;
+    }
+    foreach(PackageKitResource* res, m_packagesToDelete) {
+        const auto pkgs = m_packages.packageToApp.value(res->packageName(), {res->packageName()});
+        foreach(const auto &pkg, pkgs) {
+            m_packages.packages.take(pkg)->deleteLater();
+        }
+    }
+    m_packagesToAdd.clear();
+    m_packagesToDelete.clear();
     acquireFetching(false);
 }
 
@@ -176,33 +312,54 @@ void PackageKitBackend::transactionError(PackageKit::Transaction::Error, const Q
 
 void PackageKitBackend::packageDetails(const PackageKit::Details& details)
 {
-    QVector<AbstractResource*> resources = resourcesByPackageName(PackageKit::Daemon::packageName(details.packageId()), false);
-    foreach(AbstractResource* res, resources)
+    const QSet<AbstractResource*> resources = resourcesByPackageName(PackageKit::Daemon::packageName(details.packageId()));
+    if (resources.isEmpty())
+        qWarning() << "couldn't find package for" << details.packageId();
+
+    foreach(AbstractResource* res, resources) {
         qobject_cast<PackageKitResource*>(res)->setDetails(details);
+    }
 }
 
-QVector<AbstractResource*> PackageKitBackend::resourcesByPackageName(const QString& name, bool updating) const
+QSet<AbstractResource*> PackageKitBackend::resourcesByPackageName(const QString& name) const
 {
-    const Packages * const f = (updating ? &m_updatingPackages : &m_packages);
-    const QHash<QString, QStringList> *dictionary = &f->packageToApp;
-    const QHash<QString, AbstractResource*> *pkgs = &f->packages;
-
-    const QStringList names = dictionary->value(name, QStringList(name));
-    QVector<AbstractResource*> ret;
+    const QStringList names = m_packages.packageToApp.value(name, QStringList(name));
+    QSet<AbstractResource*> ret;
     ret.reserve(names.size());
     foreach(const QString& name, names) {
-        AbstractResource* res = pkgs->value(name);
+        AbstractResource* res = m_packages.packages.value(name);
         if (res)
             ret += res;
     }
     return ret;
 }
 
+template <typename T>
+T PackageKitBackend::resourcesByPackageNames(const QStringList &pkgnames) const
+{
+    T ret;
+    ret.reserve(pkgnames.size());
+    for(const QString &name : pkgnames) {
+        const QStringList names = m_packages.packageToApp.value(name, QStringList(name));
+        foreach(const QString& name, names) {
+            AbstractResource* res = m_packages.packages.value(name);
+            if (res)
+                ret += res;
+        }
+    }
+    return ret;
+}
+
 void PackageKitBackend::refreshDatabase()
 {
     if (!m_refresher) {
+        acquireFetching(true);
         m_refresher = PackageKit::Daemon::refreshCache(false);
-        connect(m_refresher.data(), &PackageKit::Transaction::finished, this, &PackageKitBackend::reloadPackageList);
+        connect(m_refresher.data(), &PackageKit::Transaction::finished, this, [this]() {
+            reloadPackageList();
+            acquireFetching(false);
+            delete m_refresher;
+        });
     } else {
         qWarning() << "already resetting";
     }
@@ -210,7 +367,7 @@ void PackageKitBackend::refreshDatabase()
 
 QVector<AbstractResource*> PackageKitBackend::allResources() const
 {
-    return m_packages.packages.values().toVector();
+    return containerValues<QVector<AbstractResource*>>(m_packages.packages);
 }
 
 AbstractResource* PackageKitBackend::resourceByPackageName(const QString& name) const
@@ -221,13 +378,14 @@ AbstractResource* PackageKitBackend::resourceByPackageName(const QString& name)
 
 QList<AbstractResource*> PackageKitBackend::searchPackageName(const QString& searchText)
 {
-    QList<AbstractResource*> ret;
-    Q_FOREACH (AbstractResource* res, m_packages.packages) {
-        if (res->name().contains(searchText, Qt::CaseInsensitive)) {
-            ret += res;
-        }
-    }
-    return ret;
+#ifdef NEWAPPSTREAM
+    const QList<Appstream::Component> components = m_appdata.search(searchText);
+#else
+    const QList<Appstream::Component> components = m_appdata.findComponentsByString(searchText, {});
+#endif
+    const QStringList ids = kTransform<QStringList>(components, [](const Appstream::Component& comp) { return comp.id(); });
+
+    return resourcesByPackageNames<QList<AbstractResource*>>(ids);
 }
 
 int PackageKitBackend::updatesCount() const
@@ -290,60 +448,47 @@ void PackageKitBackend::installApplication(AbstractResource* app)
     addTransaction(new PKTransaction({app}, Transaction::InstallRole));
 }
 
-void PackageKitBackend::cancelTransaction(AbstractResource* app)
-{
-    Q_FOREACH (Transaction* t, m_transactions) {
-        PKTransaction* pkt = qobject_cast<PKTransaction*>(t);
-        if (pkt->resource() == app) {
-            if (pkt->transaction()->allowCancel()) {
-                pkt->transaction()->cancel();
-            } else {
-                qWarning() << "trying to cancel a non-cancellable transaction: " << app->name();
-            }
-            break;
-        }
-    }
-}
-
 void PackageKitBackend::removeApplication(AbstractResource* app)
 {
     Q_ASSERT(!isFetching());
     addTransaction(new PKTransaction({app}, Transaction::RemoveRole));
 }
 
-QList<AbstractResource*> PackageKitBackend::upgradeablePackages() const
+QSet<AbstractResource*> PackageKitBackend::upgradeablePackages() const
 {
-    QVector<AbstractResource*> ret;
+    QSet<AbstractResource*> ret;
+    ret.reserve(m_updatesPackageId.size());
     Q_FOREACH (const QString& pkgid, m_updatesPackageId) {
         const QString pkgname = PackageKit::Daemon::packageName(pkgid);
-        const auto pkgs = resourcesByPackageName(pkgname, false);
+        const auto pkgs = resourcesByPackageName(pkgname);
         if (pkgs.isEmpty()) {
             qWarning() << "couldn't find resource for" << pkgid;
         }
-        ret += pkgs;
+        ret.unite(pkgs);
     }
-    return ret.toList();
+    return ret;
 }
 
 void PackageKitBackend::addPackageToUpdate(PackageKit::Transaction::Info info, const QString& packageId, const QString& summary)
 {
     if (info != PackageKit::Transaction::InfoBlocked) {
         m_updatesPackageId += packageId;
-        addPackage(info, packageId, summary);
+        addPackage(info, packageId, summary, true);
     }
 }
 
 void PackageKitBackend::getUpdatesFinished(PackageKit::Transaction::Exit, uint)
 {
     if (!m_updatesPackageId.isEmpty()) {
-        acquireFetching(true);
         PackageKit::Transaction* transaction = PackageKit::Daemon::getDetails(m_updatesPackageId.toList());
         connect(transaction, &PackageKit::Transaction::details, this, &PackageKitBackend::packageDetails);
         connect(transaction, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
         connect(transaction, &PackageKit::Transaction::finished, this, &PackageKitBackend::getUpdatesDetailsFinished);
     }
 
-    acquireFetching(false);
+    m_updater->setProgressing(false);
+
+    includePackagesToAdd();
     emit updatesCountChanged();
 }
 
@@ -352,7 +497,6 @@ void PackageKitBackend::getUpdatesDetailsFinished(PackageKit::Transaction::Exit
     if (exit != PackageKit::Transaction::ExitSuccess) {
         qWarning() << "Couldn't figure out the updates on PackageKit backend" << exit;
     }
-    acquireFetching(false);
 }
 
 bool PackageKitBackend::isPackageNameUpgradeable(const PackageKitResource* res) const
@@ -363,13 +507,30 @@ bool PackageKitBackend::isPackageNameUpgradeable(const PackageKitResource* res)
 QString PackageKitBackend::upgradeablePackageId(const PackageKitResource* res) const
 {
     QString name = res->packageName();
-    for (const QString& pkgid: m_updatesPackageId) {
+    foreach (const QString& pkgid, m_updatesPackageId) {
         if (PackageKit::Daemon::packageName(pkgid) == name)
             return pkgid;
     }
     return QString();
 }
 
+void PackageKitBackend::fetchDetails(const QString& pkgid)
+{
+    if (!m_delayedDetailsFetch.isActive()) {
+        m_delayedDetailsFetch.start();
+    }
+
+    m_packageNamesToFetchDetails += pkgid;
+}
+
+void PackageKitBackend::performDetailsFetch()
+{
+    Q_ASSERT(!m_packageNamesToFetchDetails.isEmpty());
+    PackageKit::Transaction* transaction = PackageKit::Daemon::getDetails(m_packageNamesToFetchDetails.toList());
+    connect(transaction, &PackageKit::Transaction::details, this, &PackageKitBackend::packageDetails);
+    connect(transaction, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
+}
+
 void PackageKitBackend::checkDaemonRunning()
 {
     if (!PackageKit::Daemon::isRunning()) {
