@@ -19,6 +19,7 @@
  ***************************************************************************/
 
 #include "ResourcesModel.h"
+#include <utils.h>
 
 #include "AbstractResource.h"
 #include "resources/AbstractResourcesBackend.h"
@@ -55,6 +56,7 @@ ResourcesModel::ResourcesModel(QObject* parent, bool load)
         { StateRole, "state" },
         { RatingRole, "rating" },
         { RatingPointsRole, "ratingPoints" },
+        { RatingCountRole, "ratingCount" },
         { SortableRatingRole, "sortableRating" },
         { ActiveRole, "active" },
         { InstalledRole, "isInstalled" },
@@ -64,8 +66,10 @@ ResourcesModel::ResourcesModel(QObject* parent, bool load)
         { PackageNameRole, "packageName" },
         { IsTechnicalRole, "isTechnical" },
         { CategoryRole, "category" },
+        { CategoryDisplayRole, "categoryDisplay" },
         { SectionRole, "section" },
         { MimeTypes, "mimetypes" },
+        { LongDescriptionRole, "longDescription" },
         { SizeRole, "size" }
         })
     )
@@ -130,6 +134,7 @@ void ResourcesModel::addResourcesBackend(AbstractResourcesBackend* backend)
 
     connect(backend, &AbstractResourcesBackend::fetchingChanged, this, &ResourcesModel::callerFetchingChanged);
     connect(backend, &AbstractResourcesBackend::allDataChanged, this, &ResourcesModel::updateCaller);
+    connect(backend, &AbstractResourcesBackend::resourcesChanged, this, &ResourcesModel::emitResourceChanges);
     connect(backend, &AbstractResourcesBackend::updatesCountChanged, this, &ResourcesModel::updatesCountChanged);
     connect(backend, &AbstractResourcesBackend::searchInvalidated, this, &ResourcesModel::searchInvalidated);
 
@@ -162,8 +167,7 @@ QModelIndex ResourcesModel::resourceIndex(AbstractResource* res) const
         else {
             Q_ASSERT(!m_backends[i]->isFetching());
             int pos = m_resources[i].indexOf(res);
-            Q_ASSERT(pos>=0);
-            return index(row+pos);
+            return pos>=0 ? index(row+pos) : QModelIndex();
         }
     }
 
@@ -176,7 +180,7 @@ QVariant ResourcesModel::data(const QModelIndex& index, int role) const
         return QVariant();
     }
 
-    AbstractResource* resource = resourceAt(index.row());
+    AbstractResource* const resource = resourceAt(index.row());
     switch(role) {
         case ActiveRole:
             return TransactionModel::global()->transactionFromResource(resource) != nullptr;
@@ -184,9 +188,17 @@ QVariant ResourcesModel::data(const QModelIndex& index, int role) const
             return qVariantFromValue<QObject*>(resource);
         case RatingPointsRole:
         case RatingRole:
+        case RatingCountRole:
         case SortableRatingRole: {
-            Rating* rating = resource->rating();
-            return rating ? rating->property(roleNames().value(role).constData()) : -1;
+            Rating* const rating = resource->rating();
+            if (rating)
+                return rating->property(roleNames().value(role).constData());
+            else {
+                const int idx = Rating::staticMetaObject.indexOfProperty(roleNames().value(role).constData());
+                QVariant val = QVariant(0);
+                val.convert(Rating::staticMetaObject.property(idx).type());
+                return val;
+            }
         }
         case Qt::DecorationRole:
         case Qt::DisplayRole:
@@ -298,7 +310,7 @@ void ResourcesModel::resetBackend(AbstractResourcesBackend* backend)
     }
 }
 
-void ResourcesModel::updateCaller()
+void ResourcesModel::updateCaller(const QVector<QByteArray>& properties)
 {
     AbstractResourcesBackend* backend = qobject_cast<AbstractResourcesBackend*>(sender());
     QVector<QVector<AbstractResource*>>::iterator backendsResources;
@@ -306,7 +318,25 @@ void ResourcesModel::updateCaller()
     if (backendsResources->isEmpty())
         return;
     
-    emit dataChanged(index(before), index(before+backendsResources->size()-1));
+    emit dataChanged(index(before), index(before+backendsResources->size()-1), propertiesToRoles(properties));
+}
+
+QVector<int> ResourcesModel::propertiesToRoles(const QVector<QByteArray>& properties) const
+{
+    QVector<int> roles = kTransform<QVector<int>>(properties, [this](const QByteArray& arr) { return roleNames().key(arr, -1); });
+    roles.removeAll(-1);
+    return roles;
+}
+
+void ResourcesModel::emitResourceChanges(AbstractResource* resource, const QVector<QByteArray> &properties)
+{
+    resourceDataChanged(resource, properties);
+
+    const QModelIndex idx = resourceIndex(resource);
+    if (!idx.isValid())
+        return;
+
+    emit dataChanged(idx, idx, propertiesToRoles(properties));
 }
 
 QVector< AbstractResourcesBackend* > ResourcesModel::backends() const
@@ -316,6 +346,9 @@ QVector< AbstractResourcesBackend* > ResourcesModel::backends() const
 
 AbstractResource* ResourcesModel::resourceByPackageName(const QString& name)
 {
+    if (m_backends.isEmpty()) {
+        qWarning() << "looking for package" << name << "without any backends loaded";
+    }
     foreach(AbstractResourcesBackend* backend, m_backends)
     {
         AbstractResource* res = backend->resourceByPackageName(name);
@@ -355,12 +388,6 @@ void ResourcesModel::removeApplication(AbstractResource* app)
     app->backend()->removeApplication(app);
 }
 
-void ResourcesModel::cancelTransaction(AbstractResource* app)
-{
-    Q_ASSERT(!isFetching());
-    app->backend()->cancelTransaction(app);
-}
-
 QMap<int, QVariant> ResourcesModel::itemData(const QModelIndex& index) const
 {
     QMap<int, QVariant> ret;
@@ -431,19 +458,19 @@ QList<QAction*> ResourcesModel::messageActions() const
     return ret;
 }
 
-QVariantList ResourcesModel::messageActionsVariant() const
-{
-    QVariantList ret;
-    QList<QAction*> actions = messageActions();
-    ret.reserve(actions.count());
-
-    foreach(QAction* action, actions) {
-        ret += QVariant::fromValue<QObject*>(action);
-    }
-    return ret;
-}
-
 bool ResourcesModel::isBusy() const
 {
     return TransactionModel::global()->rowCount() > 0;
 }
+
+bool ResourcesModel::isExtended(const QString& id)
+{
+    auto resourceExtends = [id](AbstractResource* res) {return res->extends().contains(id); };
+    bool ret = false;
+    foreach (const QVector<AbstractResource*> & backend, m_resources) {
+        ret = std::find_if(backend.cbegin(), backend.cend(), resourceExtends) != backend.cend();
+        if (ret)
+            break;
+    }
+    return ret;
+}
